import glob
import os

from cpl_core.database.context import DatabaseContextABC
from cpl_core.dependency_injection import ServiceProviderABC
from cpl_query.extension import List

from bot_core.logging.database_logger import DatabaseLogger
from bot_data.model.migration import Migration
from bot_data.model.migration_history import MigrationHistory


class MigrationService:
    def __init__(
        self,
        logger: DatabaseLogger,
        services: ServiceProviderABC,
        db: DatabaseContextABC,
    ):
        self._logger = logger
        self._services = services

        self._db = db
        self._cursor = db.cursor

        # self._migrations: List[MigrationABC] = (
        #     List(type, MigrationABC.__subclasses__()).order_by(lambda x: x.name.split("_")[0]).then_by(lambda x: x.prio)
        # )

    def _migrate_from_old_to_new(self):
        results = self._db.select(
            """
            SELECT * FROM `MigrationHistory`
        """
        )
        applied_migrations = List(MigrationHistory)
        for result in results:
            applied_migrations.add(MigrationHistory(result[0], result[1]))

        for migration in applied_migrations:
            if not migration.migration_id.endswith("Migration"):
                continue

            self._logger.debug(__name__, f"Migrate old migration {migration.migration_id} to new method")
            self._cursor.execute(migration.change_id_string(migration.migration_id.replace("Migration", "")))
            self._db.save_changes()

    def _load_up_scripts(self) -> List[Migration]:
        migrations = List(Migration)
        path = "../../src/bot_data/scripts"

        if not os.path.exists(path):
            raise Exception("Migration path not found")

        folders = List(str, glob.glob(f"{path}/*")).order_by()

        for folder in folders:
            splitted = folder.split("/")
            version = splitted[len(splitted) - 1]

            for file in List(str, os.listdir(folder)).where(lambda x: x.endswith("_up.sql")).order_by().to_list():
                if not file.endswith(".sql"):
                    continue

                name = str(file.split(".sql")[0])
                if name.endswith("_up"):
                    name = name.replace("_up", "")
                elif name.endswith("_down"):
                    name = name.replace("_down", "")

                name = name.split("_")[1]

                with open(f"{folder}/{file}", "r") as f:
                    script = f.read()
                    f.close()

                migrations.add(Migration(name, version, script))

        return migrations

    def _execute(self, migrations: List[Migration]):
        for migration in migrations:
            active_statement = ""
            try:
                # check if table exists
                self._cursor.execute("SHOW TABLES LIKE 'MigrationHistory'")
                result = self._cursor.fetchone()
                if result:
                    # there is a table named "tableName"
                    self._logger.trace(
                        __name__,
                        f"Running SQL Command: {MigrationHistory.get_select_by_id_string(migration.name)}",
                    )
                    migration_from_db = self._db.select(MigrationHistory.get_select_by_id_string(migration.name))
                    if migration_from_db is not None and len(migration_from_db) > 0:
                        continue

                self._logger.debug(__name__, f"Running migration: {migration.name}")

                for statement in migration.script.split("\n\n"):
                    if statement in ["", "\n"]:
                        continue
                    active_statement = statement
                    self._cursor.execute(statement + ";")

                self._cursor.execute(MigrationHistory(migration.name).insert_string)
                self._db.save_changes()
            except Exception as e:
                self._logger.fatal(__name__, f"Migration failed: {migration.name}\n{active_statement}", e)

    def migrate(self):
        self._migrate_from_old_to_new()
        self._execute(self._load_up_scripts())

    # def migrate(self):
    #     self._logger.info(__name__, f"Running Migrations")
    #     for migration in self._migrations:
    #         migration_id = migration.__name__
    #         try:
    #             # check if table exists
    #             self._cursor.execute("SHOW TABLES LIKE 'MigrationHistory'")
    #             result = self._cursor.fetchone()
    #             if result:
    #                 # there is a table named "tableName"
    #                 self._logger.trace(
    #                     __name__,
    #                     f"Running SQL Command: {MigrationHistory.get_select_by_id_string(migration_id)}",
    #                 )
    #                 migration_from_db = self._db.select(MigrationHistory.get_select_by_id_string(migration_id))
    #                 if migration_from_db is not None and len(migration_from_db) > 0:
    #                     continue
    #
    #             self._logger.debug(__name__, f"Running Migration {migration_id}")
    #             migration_as_service: MigrationABC = self._services.get_service(migration)
    #             migration_as_service.upgrade()
    #             self._cursor.execute(MigrationHistory(migration_id).insert_string)
    #             self._db.save_changes()
    #
    #         except Exception as e:
    #             self._logger.error(__name__, f"Cannot get migration with id {migration}", e)
